@@ -74,7 +74,7 @@ namespace tiledb {
  * Query query(ctx, array);
  * query.set_layout(TILEDB_GLOBAL_ORDER);
  * std::vector a1_data = {1, 2, 3};
- * query.set_buffer("a1", a1_data);
+ * query.set_data_buffer("a1", a1_data);
  * query.submit();
  * query.finalize();
  * array.close();
@@ -287,13 +287,13 @@ class Query {
    * flush any internal state. For the case of reads, if the returned status is
    * `TILEDB_INCOMPLETE`, TileDB could not fit the entire result in the user's
    * buffers. In this case, the user should consume the read results (if any),
-   * optionally reset the buffers with `set_buffer()`, and then
+   * optionally reset the buffers with `set_data_buffer()`, and then
    * resubmit the query until the status becomes `TILEDB_COMPLETED`. If all
    * buffer sizes after the termination of this function become 0, then this
    * means that **no** useful data was read into the buffers, implying that the
    * larger buffers are needed for the query to proceed. In this case, the users
    * must reallocate their buffers (increasing their size), reset the buffers
-   * with `set_buffer()`, and resubmit the query.
+   * with `set_data_buffer()`, and resubmit the query.
    *
    * @return Query status
    */
@@ -1107,7 +1107,6 @@ class Query {
    *
    * - `sm.memory_budget`
    * - `sm.memory_budget_var`
-   * - `sm.sub_partitioner_memory_budget`
    * - `sm.var_offsets.mode`
    * - `sm.var_offsets.extra_element`
    * - `sm.var_offsets.bitsize`
@@ -1162,7 +1161,7 @@ class Query {
   template <typename T>
   TILEDB_DEPRECATED Query& set_coordinates(T* buf, uint64_t size) {
     impl::type_check<T>(schema_.domain().type());
-    return set_buffer("__coords", buf, size);
+    return set_data_buffer("__coords", buf, size);
   }
 
   /**
@@ -1211,7 +1210,8 @@ class Query {
    * @param nelements Number of array elements
    **/
   template <typename T>
-  Query& set_buffer(const std::string& name, T* buff, uint64_t nelements) {
+  TILEDB_DEPRECATED Query& set_buffer(
+      const std::string& name, T* buff, uint64_t nelements) {
     // Checks
     auto is_attr = schema_.has_attribute(name);
     auto is_dim = schema_.domain().has_dimension(name);
@@ -1226,7 +1226,7 @@ class Query {
     else if (name == "__coords")
       impl::type_check<T>(schema_.domain().type());
 
-    return set_buffer(name, buff, nelements, sizeof(T));
+    return set_data_buffer(name, buff, nelements, sizeof(T));
   }
 
   /**
@@ -1246,8 +1246,9 @@ class Query {
    * @param buf Buffer vector with elements of the attribute/dimension type.
    **/
   template <typename T>
-  Query& set_buffer(const std::string& name, std::vector<T>& buf) {
-    return set_buffer(name, buf.data(), buf.size(), sizeof(T));
+  TILEDB_DEPRECATED Query& set_buffer(
+      const std::string& name, std::vector<T>& buf) {
+    return set_data_buffer(name, buf.data(), buf.size(), sizeof(T));
   }
 
   /**
@@ -1261,7 +1262,8 @@ class Query {
    * @param buff Buffer array pointer with elements of the attribute type.
    * @param nelements Number of array elements in buffer
    **/
-  Query& set_buffer(const std::string& name, void* buff, uint64_t nelements) {
+  TILEDB_DEPRECATED Query& set_buffer(
+      const std::string& name, void* buff, uint64_t nelements) {
     // Checks
     auto is_attr = schema_.has_attribute(name);
     auto is_dim = schema_.domain().has_dimension(name);
@@ -1280,7 +1282,7 @@ class Query {
       element_size =
           tiledb_datatype_size(schema_.domain().dimension(name).type());
 
-    return set_buffer(name, buff, nelements, element_size);
+    return set_data_buffer(name, buff, nelements, element_size);
   }
 
   /**
@@ -1310,7 +1312,7 @@ class Query {
    * @param data_nelements Number of array elements in data buffer.
    **/
   template <typename T>
-  Query& set_buffer(
+  TILEDB_DEPRECATED Query& set_buffer(
       const std::string& name,
       uint64_t* offsets,
       uint64_t offset_nelements,
@@ -1328,8 +1330,9 @@ class Query {
     else if (is_dim)
       impl::type_check<T>(schema_.domain().dimension(name).type());
 
-    return set_buffer(
-        name, offsets, offset_nelements, data, data_nelements, sizeof(T));
+    set_data_buffer(name, data, data_nelements, sizeof(T));
+    set_offsets_buffer(name, offsets, offset_nelements);
+    return *this;
   }
 
   /**
@@ -1346,7 +1349,7 @@ class Query {
    * @param data Buffer array pointer with elements of the attribute type.
    * @param data_nelements Number of array elements in data buffer.
    **/
-  Query& set_buffer(
+  TILEDB_DEPRECATED Query& set_buffer(
       const std::string& name,
       uint64_t* offsets,
       uint64_t offset_nelements,
@@ -1365,8 +1368,9 @@ class Query {
                           schema_.domain().dimension(name).type();
     size_t element_size = tiledb_datatype_size(type);
 
-    return set_buffer(
-        name, offsets, offset_nelements, data, data_nelements, element_size);
+    set_data_buffer(name, data, data_nelements, element_size);
+    set_offsets_buffer(name, offsets, offset_nelements);
+    return *this;
   }
 
   /**
@@ -1391,7 +1395,7 @@ class Query {
    *        std::string, where the values of each cell are concatenated.
    **/
   template <typename T>
-  Query& set_buffer(
+  TILEDB_DEPRECATED Query& set_buffer(
       const std::string& name,
       std::vector<uint64_t>& offsets,
       std::vector<T>& data) {
@@ -1407,8 +1411,9 @@ class Query {
     else if (is_dim)
       impl::type_check<T>(schema_.domain().dimension(name).type());
 
-    return set_buffer(
-        name, offsets.data(), offsets.size(), &data[0], data.size(), sizeof(T));
+    set_data_buffer(name, data.data(), data.size(), sizeof(T));
+    set_offsets_buffer(name, offsets.data(), offsets.size());
+    return *this;
   }
 
   /**
@@ -1419,7 +1424,7 @@ class Query {
    * @param buf Pair of offset, data buffers
    **/
   template <typename T>
-  Query& set_buffer(
+  TILEDB_DEPRECATED Query& set_buffer(
       const std::string& name,
       std::pair<std::vector<uint64_t>, std::vector<T>>& buf) {
     // Checks
@@ -1434,7 +1439,9 @@ class Query {
     else if (is_dim)
       impl::type_check<T>(schema_.domain().dimension(name).type());
 
-    return set_buffer(name, buf.first, buf.second);
+    set_data_buffer(name, buf.second);
+    set_offsets_buffer(name, buf.first);
+    return *this;
   }
 
   /**
@@ -1444,7 +1451,7 @@ class Query {
    * @param offsets Offsets where a new element begins in the data buffer.
    * @param data Pre-allocated string buffer.
    **/
-  Query& set_buffer(
+  TILEDB_DEPRECATED Query& set_buffer(
       const std::string& name,
       std::vector<uint64_t>& offsets,
       std::string& data) {
@@ -1460,13 +1467,252 @@ class Query {
     else if (is_dim)
       impl::type_check<char>(schema_.domain().dimension(name).type());
 
-    return set_buffer(
-        name,
-        offsets.data(),
-        offsets.size(),
-        &data[0],
-        data.size(),
-        sizeof(char));
+    set_data_buffer(name, &data[0], data.size(), sizeof(char));
+    set_offsets_buffer(name, offsets.data(), offsets.size());
+    return *this;
+  }
+
+  /**
+   * Sets the data for a fixed/var-sized attribute/dimension.
+   *
+   * **Example:**
+   * @code{.cpp}
+   * tiledb::Context ctx;
+   * tiledb::Array array(ctx, array_name, TILEDB_WRITE);
+   * int data_a1[] = {0, 1, 2, 3};
+   * Query query(ctx, array);
+   * query.set_data_buffer("a1", data_a1, 4);
+   * @endcode
+   *
+   * @note set_data_buffer(std::string, std::vector) is preferred as it is
+   * safer.
+   *
+   * @tparam T Attribute/Dimension value type
+   * @param name Attribute/Dimension name
+   * @param buff Buffer array pointer with elements of the
+   *     attribute/dimension type.
+   * @param nelements Number of array elements
+   **/
+  template <typename T>
+  Query& set_data_buffer(const std::string& name, T* buff, uint64_t nelements) {
+    // Checks
+    auto is_attr = schema_.has_attribute(name);
+    auto is_dim = schema_.domain().has_dimension(name);
+    if (name != "__coords" && !is_attr && !is_dim)
+      throw TileDBError(
+          std::string("Cannot set buffer; Attribute/Dimension '") + name +
+          "' does not exist");
+    else if (is_attr)
+      impl::type_check<T>(schema_.attribute(name).type());
+    else if (is_dim)
+      impl::type_check<T>(schema_.domain().dimension(name).type());
+    else if (name == "__coords")
+      impl::type_check<T>(schema_.domain().type());
+
+    return set_data_buffer(name, buff, nelements, sizeof(T));
+  }
+
+  /**
+   * Sets the data for a fixed/var-sized attribute/dimension.
+   *
+   * **Example:**
+   * @code{.cpp}
+   * tiledb::Context ctx;
+   * tiledb::Array array(ctx, array_name, TILEDB_WRITE);
+   * std::vector<int> data_a1 = {0, 1, 2, 3};
+   * Query query(ctx, array);
+   * query.set_data_buffer("a1", data_a1);
+   * @endcode
+   *
+   * @tparam T Attribute/Dimension value type
+   * @param name Attribute/Dimension name
+   * @param buf Buffer vector with elements of the attribute/dimension type.
+   **/
+  template <typename T>
+  Query& set_data_buffer(const std::string& name, std::vector<T>& buf) {
+    return set_data_buffer(name, buf.data(), buf.size(), sizeof(T));
+  }
+
+  /**
+   * Sets the data for a fixed/var-sized attribute/dimension.
+   *
+   * @note This unsafe version does not perform type checking; the given buffer
+   * is assumed to be the correct type, and the size of an element in the given
+   * buffer is assumed to be the size of the datatype of the attribute.
+   *
+   * @param name Attribute/Dimension name
+   * @param buff Buffer array pointer with elements of the attribute type.
+   * @param nelements Number of array elements in buffer
+   **/
+  Query& set_data_buffer(
+      const std::string& name, void* buff, uint64_t nelements) {
+    // Checks
+    auto is_attr = schema_.has_attribute(name);
+    auto is_dim = schema_.domain().has_dimension(name);
+    if (name != "__coords" && !is_attr && !is_dim)
+      throw TileDBError(
+          std::string("Cannot set buffer; Attribute/Dimension '") + name +
+          "' does not exist");
+
+    // Compute element size (in bytes).
+    size_t element_size = 0;
+    if (name == "__coords")
+      element_size = tiledb_datatype_size(schema_.domain().type());
+    else if (is_attr)
+      element_size = tiledb_datatype_size(schema_.attribute(name).type());
+    else if (is_dim)
+      element_size =
+          tiledb_datatype_size(schema_.domain().dimension(name).type());
+
+    return set_data_buffer(name, buff, nelements, element_size);
+  }
+
+  /**
+   * Sets the data for a fixed/var-sized attribute/dimension.
+   *
+   * @param name Attribute/Dimension name
+   * @param data Pre-allocated string buffer.
+   **/
+  Query& set_data_buffer(const std::string& name, std::string& data) {
+    // Checks
+    auto is_attr = schema_.has_attribute(name);
+    auto is_dim = schema_.domain().has_dimension(name);
+    if (!is_attr && !is_dim)
+      throw TileDBError(
+          std::string("Cannot set buffer; Attribute/Dimension '") + name +
+          "' does not exist");
+    else if (is_attr)
+      impl::type_check<char>(schema_.attribute(name).type());
+    else if (is_dim)
+      impl::type_check<char>(schema_.domain().dimension(name).type());
+
+    return set_data_buffer(name, &data[0], data.size(), sizeof(char));
+  }
+
+  /**
+   * Sets the offset buffer for a var-sized attribute/dimension.
+   *
+   * **Example:**
+   *
+   * @code{.cpp}
+   * tiledb::Context ctx;
+   * tiledb::Array array(ctx, array_name, TILEDB_WRITE);
+   * uint64_t offsets_a1[] = {0, 8};
+   * Query query(ctx, array);
+   * query.set_offsets_buffer("a1", offsets_a1, 2);
+   * @endcode
+   *
+   * @note set_offsets_buffer(std::string, std::vector, std::vector) is
+   * preferred as it is safer.
+   *
+   * @param attr Attribute/Dimension name
+   * @param offsets Offsets array pointer where a new element begins in the data
+   *        buffer.
+   * @param offsets_nelements Number of elements in offsets buffer.
+   **/
+  Query& set_offsets_buffer(
+      const std::string& attr, uint64_t* offsets, uint64_t offset_nelements) {
+    auto ctx = ctx_.get();
+    auto offset_size = offset_nelements * sizeof(uint64_t);
+    auto buff_sizes_iter = buff_sizes_.find(attr);
+    if (buff_sizes_iter == buff_sizes_.end()) {
+      buff_sizes_[attr] =
+          std::tuple<uint64_t, uint64_t, uint64_t>(offset_size, 0, 0);
+    } else {
+      auto& second = buff_sizes_iter->second;
+      buff_sizes_[attr] = std::tuple<uint64_t, uint64_t, uint64_t>(
+          offset_size, std::get<1>(second), std::get<2>(second));
+    }
+
+    ctx.handle_error(tiledb_query_set_offsets_buffer(
+        ctx.ptr().get(),
+        query_.get(),
+        attr.c_str(),
+        offsets,
+        &std::get<0>(buff_sizes_[attr])));
+    return *this;
+  }
+
+  /**
+   * Sets the offset buffer for a var-sized attribute/dimension.
+   *
+   * **Example:**
+   * @code{.cpp}
+   * tiledb::Context ctx;
+   * tiledb::Array array(ctx, array_name, TILEDB_WRITE);
+   * std::vector<uint64_t> offsets_a1 = {0, 8};
+   * Query query(ctx, array);
+   * query.set_offsets_buffer("a1", offsets_a1);
+   * @endcode
+   *
+   * @param name Attribute/Dimension name
+   * @param offsets Offsets where a new element begins in the data buffer.
+   **/
+  Query& set_offsets_buffer(
+      const std::string& name, std::vector<uint64_t>& offsets) {
+    return set_offsets_buffer(name, offsets.data(), offsets.size());
+  }
+
+  /**
+   * Sets the validity buffer for nullable attribute/dimension.
+   *
+   * @tparam T Attribute value type
+   * @param attr Attribute name
+   * @param validity_bytemap The validity bytemap buffer.
+   * @param validity_bytemap_nelements The number of values within
+   *     `validity_bytemap_nelements`
+   **/
+  Query& set_validity_buffer(
+      const std::string& attr,
+      uint8_t* validity_bytemap,
+      uint64_t validity_bytemap_nelements) {
+    auto ctx = ctx_.get();
+    size_t validity_size = validity_bytemap_nelements * sizeof(uint8_t);
+    auto buff_sizes_iter = buff_sizes_.find(attr);
+    if (buff_sizes_iter == buff_sizes_.end()) {
+      buff_sizes_[attr] =
+          std::tuple<uint64_t, uint64_t, uint64_t>(0, 0, validity_size);
+    } else {
+      auto& second = buff_sizes_iter->second;
+      buff_sizes_[attr] = std::tuple<uint64_t, uint64_t, uint64_t>(
+          std::get<0>(second), std::get<1>(second), validity_size);
+    }
+
+    ctx.handle_error(tiledb_query_set_validity_buffer(
+        ctx.ptr().get(),
+        query_.get(),
+        attr.c_str(),
+        validity_bytemap,
+        &std::get<2>(buff_sizes_[attr])));
+    return *this;
+  }
+
+  /**
+   * Sets the validity buffer for nullable attribute/dimension.
+   *
+   * **Example:**
+   * @code{.cpp}
+   * tiledb::Context ctx;
+   * tiledb::Array array(ctx, array_name, TILEDB_WRITE);
+   * std::vector<uint8_t> validity_bytemap = {1, 1, 0, 1};
+   * Query query(ctx, array);
+   * query.set_validity_buffer("a1", validity_bytemap);
+   * @endcode
+   *
+   * @param name Attribute name
+   * @param validity_bytemap Buffer vector with elements of the attribute
+   *     validity values.
+   **/
+  Query& set_validity_buffer(
+      const std::string& name, std::vector<uint8_t>& validity_bytemap) {
+    // Checks
+    auto is_attr = schema_.has_attribute(name);
+    if (!is_attr)
+      throw TileDBError(
+          std::string("Cannot set buffer; Attribute '") + name +
+          "' does not exist");
+    return set_validity_buffer(
+        name, validity_bytemap.data(), validity_bytemap.size());
   }
 
   /**
@@ -1495,7 +1741,7 @@ class Query {
    *     `validity_bytemap_nelements`
    **/
   template <typename T>
-  Query& set_buffer_nullable(
+  TILEDB_DEPRECATED Query& set_buffer_nullable(
       const std::string& name,
       T* data,
       uint64_t data_nelements,
@@ -1510,13 +1756,9 @@ class Query {
     else
       impl::type_check<T>(schema_.attribute(name).type());
 
-    return set_buffer_nullable(
-        name,
-        data,
-        data_nelements,
-        sizeof(T),
-        validity_bytemap,
-        validity_bytemap_nelements);
+    set_data_buffer(name, data, data_nelements);
+    set_validity_buffer(name, validity_bytemap, validity_bytemap_nelements);
+    return *this;
   }
 
   /**
@@ -1539,17 +1781,13 @@ class Query {
    *     validity values.
    **/
   template <typename T>
-  Query& set_buffer_nullable(
+  TILEDB_DEPRECATED Query& set_buffer_nullable(
       const std::string& name,
       std::vector<T>& buf,
       std::vector<uint8_t>& validity_bytemap) {
-    return set_buffer_nullable(
-        name,
-        buf.data(),
-        buf.size(),
-        sizeof(T),
-        validity_bytemap.data(),
-        validity_bytemap.size());
+    set_data_buffer(name, buf.data(), buf.size(), sizeof(T));
+    set_validity_buffer(name, validity_bytemap.data(), validity_bytemap.size());
+    return *this;
   }
 
   /**
@@ -1566,7 +1804,7 @@ class Query {
    * @param validity_bytemap_nelements The number of values within
    *     `validity_bytemap_nelements`
    **/
-  Query& set_buffer_nullable(
+  TILEDB_DEPRECATED Query& set_buffer_nullable(
       const std::string& name,
       void* data,
       uint64_t data_nelements,
@@ -1582,13 +1820,9 @@ class Query {
     // Compute element size (in bytes).
     size_t element_size = tiledb_datatype_size(schema_.attribute(name).type());
 
-    return set_buffer_nullable(
-        name,
-        data,
-        data_nelements,
-        element_size,
-        validity_bytemap,
-        validity_bytemap_nelements);
+    set_data_buffer(name, data, data_nelements, element_size);
+    set_validity_buffer(name, validity_bytemap, validity_bytemap_nelements);
+    return *this;
   }
 
   /**
@@ -1608,7 +1842,7 @@ class Query {
    * @param validity_bytemap_nelements The number of values within
    *     `validity_bytemap_nelements`
    **/
-  Query& set_buffer_nullable(
+  TILEDB_DEPRECATED Query& set_buffer_nullable(
       const std::string& name,
       uint64_t* offsets,
       uint64_t offset_nelements,
@@ -1627,15 +1861,10 @@ class Query {
     auto type = schema_.attribute(name).type();
     size_t element_size = tiledb_datatype_size(type);
 
-    return set_buffer_nullable(
-        name,
-        offsets,
-        offset_nelements,
-        data,
-        data_nelements,
-        element_size,
-        validity_bytemap,
-        validity_bytemap_nelements);
+    set_data_buffer(name, data, data_nelements, element_size);
+    set_offsets_buffer(name, offsets, offset_nelements);
+    set_validity_buffer(name, validity_bytemap, validity_bytemap_nelements);
+    return *this;
   }
 
   /**
@@ -1663,7 +1892,7 @@ class Query {
    *     validity values.
    **/
   template <typename T>
-  Query& set_buffer_nullable(
+  TILEDB_DEPRECATED Query& set_buffer_nullable(
       const std::string& name,
       std::vector<uint64_t>& offsets,
       std::vector<T>& data,
@@ -1677,15 +1906,10 @@ class Query {
     else
       impl::type_check<T>(schema_.attribute(name).type());
 
-    return set_buffer_nullable(
-        name,
-        offsets.data(),
-        offsets.size(),
-        data.data(),
-        data.size(),
-        sizeof(T),
-        validity_bytemap.data(),
-        validity_bytemap.size());
+    set_data_buffer(name, data.data(), data.size(), sizeof(T));
+    set_offsets_buffer(name, offsets.data(), offsets.size());
+    set_validity_buffer(name, validity_bytemap.data(), validity_bytemap.size());
+    return *this;
   }
 
   /**
@@ -1696,7 +1920,7 @@ class Query {
    * @param buf Pair of offset, data, validity bytemap buffers
    **/
   template <typename T>
-  Query& set_buffer_nullable(
+  TILEDB_DEPRECATED Query& set_buffer_nullable(
       const std::string& name,
       std::tuple<std::vector<uint64_t>, std::vector<T>, std::vector<uint8_t>>&
           buf) {
@@ -1708,8 +1932,10 @@ class Query {
           "' does not exist");
     impl::type_check<T>(schema_.attribute(name).type());
 
-    return set_buffer_nullable(
-        name, std::get<0>(buf), std::get<1>(buf), std::get<2>(buf));
+    set_data_buffer(name, std::get<1>(buf));
+    set_offsets_buffer(name, std::get<0>(buf));
+    set_validity_buffer(name, std::get<2>(buf));
+    return *this;
   }
 
   /**
@@ -1721,7 +1947,7 @@ class Query {
    * @param validity_bytemap Buffer vector with elements of the attribute
    *     validity values.
    **/
-  Query& set_buffer_nullable(
+  TILEDB_DEPRECATED Query& set_buffer_nullable(
       const std::string& name,
       std::vector<uint64_t>& offsets,
       std::string& data,
@@ -1735,15 +1961,10 @@ class Query {
     else
       impl::type_check<char>(schema_.attribute(name).type());
 
-    return set_buffer_nullable(
-        name,
-        offsets.data(),
-        offsets.size(),
-        &data[0],
-        data.size(),
-        sizeof(char),
-        validity_bytemap.data(),
-        validity_bytemap.size());
+    set_data_buffer(name, &data[0], data.size(), sizeof(char));
+    set_offsets_buffer(name, offsets.data(), offsets.size());
+    set_validity_buffer(name, validity_bytemap.data(), validity_bytemap.size());
+    return *this;
   }
 
   /**
@@ -1754,7 +1975,7 @@ class Query {
    * @param data_nelements Number of array elements.
    * @param element_size Size of array elements (in bytes).
    **/
-  Query& get_buffer(
+  TILEDB_DEPRECATED Query& get_buffer(
       const std::string& name,
       void** data,
       uint64_t* data_nelements,
@@ -1768,7 +1989,7 @@ class Query {
           "'!");
     }
 
-    ctx.handle_error(tiledb_query_get_buffer(
+    ctx.handle_error(tiledb_query_get_data_buffer(
         ctx.ptr().get(), query_.get(), name.c_str(), data, &data_nbytes));
 
     assert(*data_nbytes % elem_size_iter->second == 0);
@@ -1789,7 +2010,7 @@ class Query {
    * @param data_nelements Number of array elements.
    * @param element_size Size of array elements (in bytes).
    **/
-  Query& get_buffer(
+  TILEDB_DEPRECATED Query& get_buffer(
       const std::string& name,
       uint64_t** offsets,
       uint64_t* offsets_nelements,
@@ -1806,14 +2027,10 @@ class Query {
           "'!");
     }
 
-    ctx.handle_error(tiledb_query_get_buffer_var(
-        ctx.ptr().get(),
-        query_.get(),
-        name.c_str(),
-        offsets,
-        &offsets_nbytes,
-        data,
-        &data_nbytes));
+    ctx.handle_error(tiledb_query_get_data_buffer(
+        ctx.ptr().get(), query_.get(), name.c_str(), data, &data_nbytes));
+    ctx.handle_error(tiledb_query_get_offsets_buffer(
+        ctx.ptr().get(), query_.get(), name.c_str(), offsets, &offsets_nbytes));
 
     assert(*data_nbytes % elem_size_iter->second == 0);
     assert(*offsets_nbytes % sizeof(uint64_t) == 0);
@@ -1825,6 +2042,87 @@ class Query {
     return *this;
   }
 
+  /**
+   * Retrieves the data buffer of a fixed/var-sized attribute/dimension.
+   *
+   * @param name Attribute/dimension name
+   * @param data Buffer array pointer with elements of the attribute type.
+   * @param data_nelements Number of array elements.
+   * @param element_size Size of array elements (in bytes).
+   **/
+  Query& get_data_buffer(
+      const std::string& name,
+      void** data,
+      uint64_t* data_nelements,
+      uint64_t* element_size) {
+    auto ctx = ctx_.get();
+    uint64_t* data_nbytes = nullptr;
+    auto elem_size_iter = element_sizes_.find(name);
+    if (elem_size_iter == element_sizes_.end()) {
+      throw TileDBError(
+          "[TileDB::C++API] Error: No buffer set for attribute '" + name +
+          "'!");
+    }
+
+    ctx.handle_error(tiledb_query_get_data_buffer(
+        ctx.ptr().get(), query_.get(), name.c_str(), data, &data_nbytes));
+
+    assert(*data_nbytes % elem_size_iter->second == 0);
+
+    *data_nelements = (*data_nbytes) / elem_size_iter->second;
+    *element_size = elem_size_iter->second;
+
+    return *this;
+  }
+
+  /**
+   * Retrieves the offset buffer for a var-sized attribute/dimension.
+   *
+   * @param name Attribute/dimension name
+   * @param offsets Offsets array pointer with elements of uint64_t type.
+   * @param offsets_nelements Number of array elements.
+   **/
+  Query& get_offsets_buffer(
+      const std::string& name,
+      uint64_t** offsets,
+      uint64_t* offsets_nelements) {
+    auto ctx = ctx_.get();
+    uint64_t* offsets_nbytes = nullptr;
+    ctx.handle_error(tiledb_query_get_offsets_buffer(
+        ctx.ptr().get(), query_.get(), name.c_str(), offsets, &offsets_nbytes));
+
+    *offsets_nelements = (*offsets_nbytes) / sizeof(uint64_t);
+
+    return *this;
+  }
+
+  /**
+   * Retrieves the validity buffer for a nullable attribute/dimension.
+   *
+   * @param name Attribute name
+   * @param validity_bytemap Buffer array pointer with elements of the
+   *     attribute validity values.
+   * @param validity_bytemap_nelements Number of validity bytemap elements.
+   **/
+  Query& get_validity_buffer(
+      const std::string& name,
+      uint8_t** validity_bytemap,
+      uint64_t* validity_bytemap_nelements) {
+    auto ctx = ctx_.get();
+    uint64_t* validity_bytemap_nbytes = nullptr;
+
+    ctx.handle_error(tiledb_query_get_validity_buffer(
+        ctx.ptr().get(),
+        query_.get(),
+        name.c_str(),
+        validity_bytemap,
+        &validity_bytemap_nbytes));
+
+    *validity_bytemap_nelements = *validity_bytemap_nbytes / sizeof(uint8_t);
+
+    return *this;
+  }
+
   /**
    * Gets a buffer for a fixed-sized, nullable attribute.
    *
@@ -1836,7 +2134,7 @@ class Query {
    *     attribute validity values.
    * @param validity_bytemap_nelements Number of validity bytemap elements.
    **/
-  Query& get_buffer_nullable(
+  TILEDB_DEPRECATED Query& get_buffer_nullable(
       const std::string& name,
       void** data,
       uint64_t* data_nelements,
@@ -1853,12 +2151,12 @@ class Query {
           "'!");
     }
 
-    ctx.handle_error(tiledb_query_get_buffer_nullable(
+    ctx.handle_error(tiledb_query_get_data_buffer(
+        ctx.ptr().get(), query_.get(), name.c_str(), data, &data_nbytes));
+    ctx.handle_error(tiledb_query_get_validity_buffer(
         ctx.ptr().get(),
         query_.get(),
         name.c_str(),
-        data,
-        &data_nbytes,
         validity_bytemap,
         &validity_bytemap_nbytes));
 
@@ -1884,7 +2182,7 @@ class Query {
    *     attribute validity values.
    * @param validity_bytemap_nelements Number of validity bytemap elements.
    **/
-  Query& get_buffer_nullable(
+  TILEDB_DEPRECATED Query& get_buffer_nullable(
       const std::string& name,
       uint64_t** offsets,
       uint64_t* offsets_nelements,
@@ -1904,14 +2202,14 @@ class Query {
           "'!");
     }
 
-    ctx.handle_error(tiledb_query_get_buffer_var_nullable(
+    ctx.handle_error(tiledb_query_get_data_buffer(
+        ctx.ptr().get(), query_.get(), name.c_str(), data, &data_nbytes));
+    ctx.handle_error(tiledb_query_get_offsets_buffer(
+        ctx.ptr().get(), query_.get(), name.c_str(), offsets, &offsets_nbytes));
+    ctx.handle_error(tiledb_query_get_validity_buffer(
         ctx.ptr().get(),
         query_.get(),
         name.c_str(),
-        offsets,
-        &offsets_nbytes,
-        data,
-        &data_nbytes,
         validity_bytemap,
         &validity_bytemap_nbytes));
 
@@ -2026,7 +2324,7 @@ class Query {
    * @param nelements Number of array elements.
    * @param element_size Size of array elements (in bytes).
    **/
-  Query& set_buffer(
+  TILEDB_DEPRECATED Query& set_buffer(
       const std::string& attr,
       void* buff,
       uint64_t nelements,
@@ -2035,7 +2333,7 @@ class Query {
     size_t size = nelements * element_size;
     buff_sizes_[attr] = std::tuple<uint64_t, uint64_t, uint64_t>(0, size, 0);
     element_sizes_[attr] = element_size;
-    ctx.handle_error(tiledb_query_set_buffer(
+    ctx.handle_error(tiledb_query_set_data_buffer(
         ctx.ptr().get(),
         query_.get(),
         attr.c_str(),
@@ -2057,7 +2355,7 @@ class Query {
    * @param data_nelements Number of array elements in data buffer.
    * @param data_element_size Size of data array elements (in bytes).
    **/
-  Query& set_buffer(
+  TILEDB_DEPRECATED Query& set_buffer(
       const std::string& attr,
       uint64_t* offsets,
       uint64_t offset_nelements,
@@ -2070,12 +2368,51 @@ class Query {
     element_sizes_[attr] = data_element_size;
     buff_sizes_[attr] =
         std::tuple<uint64_t, uint64_t, uint64_t>(offset_size, data_size, 0);
-    ctx.handle_error(tiledb_query_set_buffer_var(
+    ctx.handle_error(tiledb_query_set_data_buffer(
+        ctx.ptr().get(),
+        query_.get(),
+        attr.c_str(),
+        data,
+        &std::get<1>(buff_sizes_[attr])));
+    ctx.handle_error(tiledb_query_set_offsets_buffer(
         ctx.ptr().get(),
         query_.get(),
         attr.c_str(),
         offsets,
-        &std::get<0>(buff_sizes_[attr]),
+        &std::get<0>(buff_sizes_[attr])));
+    return *this;
+  }
+
+  /**
+   * Sets the data for a fixed/var-sized attribute/dimension.
+   *
+   * @param attr Attribute name
+   * @param data Buffer array pointer with elements of the attribute type.
+   * @param data_nelements Number of array elements.
+   * @param data_element_size Size of array elements (in bytes).
+   **/
+  Query& set_data_buffer(
+      const std::string& attr,
+      void* data,
+      uint64_t data_nelements,
+      size_t data_element_size) {
+    auto ctx = ctx_.get();
+    auto data_size = data_nelements * data_element_size;
+    element_sizes_[attr] = data_element_size;
+    auto buff_sizes_iter = buff_sizes_.find(attr);
+    if (buff_sizes_iter == buff_sizes_.end()) {
+      buff_sizes_[attr] =
+          std::tuple<uint64_t, uint64_t, uint64_t>(0, data_size, 0);
+    } else {
+      auto& second = buff_sizes_iter->second;
+      buff_sizes_[attr] = std::tuple<uint64_t, uint64_t, uint64_t>(
+          std::get<0>(second), data_size, std::get<2>(second));
+    }
+
+    ctx.handle_error(tiledb_query_set_data_buffer(
+        ctx.ptr().get(),
+        query_.get(),
+        attr.c_str(),
         data,
         &std::get<1>(buff_sizes_[attr])));
     return *this;
@@ -2091,7 +2428,7 @@ class Query {
    * @param validity_bytemap Buffer array pointer with validity bytemap values.
    * @param validity_bytemap_nelements Number of validity bytemap elements.
    **/
-  Query& set_buffer_nullable(
+  TILEDB_DEPRECATED Query& set_buffer_nullable(
       const std::string& attr,
       void* data,
       uint64_t data_nelements,
@@ -2104,12 +2441,16 @@ class Query {
     buff_sizes_[attr] =
         std::tuple<uint64_t, uint64_t, uint64_t>(0, data_size, validity_size);
     element_sizes_[attr] = data_element_size;
-    ctx.handle_error(tiledb_query_set_buffer_nullable(
+    ctx.handle_error(tiledb_query_set_data_buffer(
         ctx.ptr().get(),
         query_.get(),
         attr.c_str(),
         data,
-        &std::get<1>(buff_sizes_[attr]),
+        &std::get<1>(buff_sizes_[attr])));
+    ctx.handle_error(tiledb_query_set_validity_buffer(
+        ctx.ptr().get(),
+        query_.get(),
+        attr.c_str(),
         validity_bytemap,
         &std::get<2>(buff_sizes_[attr])));
     return *this;
@@ -2130,7 +2471,7 @@ class Query {
    * @param validity_bytemap Buffer array pointer with validity bytemap values.
    * @param validity_bytemap_nelements Number of validity bytemap elements.
    **/
-  Query& set_buffer_nullable(
+  TILEDB_DEPRECATED Query& set_buffer_nullable(
       const std::string& attr,
       uint64_t* offsets,
       uint64_t offset_nelements,
@@ -2146,14 +2487,22 @@ class Query {
     element_sizes_[attr] = data_element_size;
     buff_sizes_[attr] = std::tuple<uint64_t, uint64_t, uint64_t>(
         offset_size, data_size, validity_size);
-    ctx.handle_error(tiledb_query_set_buffer_var_nullable(
+    ctx.handle_error(tiledb_query_set_data_buffer(
+        ctx.ptr().get(),
+        query_.get(),
+        attr.c_str(),
+        data,
+        &std::get<1>(buff_sizes_[attr])));
+    ctx.handle_error(tiledb_query_set_offsets_buffer(
         ctx.ptr().get(),
         query_.get(),
         attr.c_str(),
         offsets,
-        &std::get<0>(buff_sizes_[attr]),
-        data,
-        &std::get<1>(buff_sizes_[attr]),
+        &std::get<0>(buff_sizes_[attr])));
+    ctx.handle_error(tiledb_query_set_validity_buffer(
+        ctx.ptr().get(),
+        query_.get(),
+        attr.c_str(),
         validity_bytemap,
         &std::get<2>(buff_sizes_[attr])));
     return *this;
